import json
import logging
from http import HTTPStatus
from typing import Callable, TYPE_CHECKING

from django.core.exceptions import ValidationError as DjangoValidationError
from django.http import Http404, HttpRequest, JsonResponse, QueryDict
from django.utils.translation import gettext

from esteid.exceptions import CanceledByUser, EsteidError, InvalidParameters


try:
    from rest_framework.exceptions import ValidationError as DRFValidationError
except ImportError:
    # If rest framework is not installed, create a stub class so the isinstance check is always false
    class DRFValidationError:
        pass


if TYPE_CHECKING:
    # Make type checkers aware of request.session attribute which is missing on the HttpRequest class
    from django.contrib.sessions import base_session

    class RequestType(HttpRequest):
        session: base_session.AbstractBaseSession
        data: dict


logger = logging.getLogger(__name__)


class SessionViewMixin:
    """
    Provides POST and PATCH method handlers for auth/signing session management.

    Also does common error handling.
    """

    class Status:
        ERROR = "error"
        PENDING = "pending"
        SUCCESS = "success"
        CANCELLED = "cancelled"

    start_session: Callable
    finish_session: Callable

    def report_error(self, e: EsteidError):
        return JsonResponse({"status": self.Status.ERROR, **e.get_user_error()}, status=e.status)

    def handle_user_cancel(self):
        pass

    def handle_errors(self, e: Exception, stage="start"):
        if isinstance(e, EsteidError):
            if isinstance(e, CanceledByUser):
                self.handle_user_cancel()

            # Do not log user input related errors
            if not isinstance(e, InvalidParameters):
                logger.exception("Failed to %s signing session.", stage)
            return self.report_error(e)

        if isinstance(e, (Http404, DRFValidationError)):
            raise e

        if isinstance(e, DjangoValidationError):
            return JsonResponse(
                {"status": self.Status.ERROR, "error": e.__class__.__name__, "message": str(e)},
                status=HTTPStatus.CONFLICT,
            )

        logger.exception("Failed to %s signing session.", stage)
        return JsonResponse(
            {"status": self.Status.ERROR, "error": "Internal error", "message": gettext("Internal server error")},
            status=HTTPStatus.INTERNAL_SERVER_ERROR,
        )

    def post(self, request, *args, **kwargs):
        """
        Handles session start requests
        """
        try:
            return self.start_session(request, *args, **kwargs)
        except Exception as e:
            return self.handle_errors(e, stage="start")

    def patch(self, request, *args, **kwargs):
        """
        Handles session finish requests
        """
        try:
            return self.finish_session(request, *args, **kwargs)
        except Exception as e:
            return self.handle_errors(e, stage="finish")


class DjangoRestCompatibilityMixin(SessionViewMixin):
    """
    Enables Django view to accept JSON request bodies the same way as Rest Framework views.
    """

    def post(self, request: "RequestType", *args, **kwargs):
        """
        Handles session start requests
        """
        request.data = self.parse_request(request)
        return super().post(request, *args, **kwargs)

    def patch(self, request: "RequestType", *args, **kwargs):
        """
        Handles session finish requests
        """
        request.data = self.parse_request(request)
        return super().patch(request, *args, **kwargs)

    @staticmethod
    def parse_request(request):
        """
        Parses PATCH/POST request bodies as JSON or urlencoded, and assigns `request.data`.

        Rationale:
        * Compatibility with REST Framework.
        * Allow JSON.
        * Django's request.POST only works for POST, not PATCH etc.
        """
        try:
            if request.content_type == "application/x-www-form-urlencoded":
                return QueryDict(request.body).dict()
            if request.content_type == "application/json":
                data = json.loads(request.body)
                if isinstance(data, dict):
                    return data
                raise InvalidParameters("Failed to parse request data as dict")
        except InvalidParameters:
            raise
        except Exception as e:
            raise InvalidParameters(
                f"Failed to parse the request body according to content type {request.content_type}"
            ) from e
        raise InvalidParameters(f"Unsupported request content type {request.content_type}")



__all__ = [
    "AuthenticationViewDjangoMixin",
    "AuthenticationViewMixin",
    "AuthenticationViewRestMixin",
    "Authenticator",
]

from .authenticator import Authenticator
from .views import AuthenticationViewDjangoMixin, AuthenticationViewMixin, AuthenticationViewRestMixin



__all__ = ["MobileIdAuthenticator", "MobileIDError", "MobileIdSigner"]

from ..exceptions import EsteidError as MobileIDError
from .authenticator import MobileIdAuthenticator
from .signer import MobileIdSigner



import os

from django.conf import settings
from django.conf.urls.static import static
from django.views.generic import TemplateView

import esteid.flowtest.urls


try:
    from django.urls import include, re_path
except ImportError:  # noqa
    from django.conf.urls import include, url as re_path

from .views import (
    SKTestView,
    TestDownloadContainerView,
    TestIdCardFinishView,
    TestIdCardSignView,
    TestMobileIdSignView,
    TestMobileIdStatusView,
    TestSmartIdSignView,
    TestSmartIdStatusView,
)


urlpatterns = []

# In a project, the URLs can be included if settings.DEBUG is on.
# In tests, settings.DEBUG is False for unknown reasons, so we check that the root urlconf points at this file.
if settings.DEBUG or settings.ROOT_URLCONF == __name__:
    urlpatterns += [
        re_path("^$", SKTestView.as_view(), name="sk_test"),
        # NOTE: compare the template test-new.html to test.html locally to see JS changes.
        re_path(r"^new/", SKTestView.as_view(template_name="esteid/test-new.html"), name="sk_test_new"),
        re_path(r"^new-auth/", TemplateView.as_view(template_name="esteid/auth-new.html"), name="sk_test_auth_new"),
        re_path(r"^download/", TestDownloadContainerView.as_view(), name="download_signed_container"),
        re_path(r"^id/start/", TestIdCardSignView.as_view(), name="test_id_start"),
        re_path(r"^id/finish/", TestIdCardFinishView.as_view(), name="test_id_finish"),
        re_path(r"^mid/start/", TestMobileIdSignView.as_view(), name="test_mid_start"),
        re_path(r"^mid/status/", TestMobileIdStatusView.as_view(), name="test_mid_status"),
        re_path(r"^smartid/start/", TestSmartIdSignView.as_view(), name="test_smartid_start"),
        re_path(r"^smartid/status/", TestSmartIdStatusView.as_view(), name="test_smartid_status"),
        re_path(r"^flowtest/", include(esteid.flowtest.urls)),
    ]

    urlpatterns += static(settings.STATIC_URL, document_root=os.path.dirname(__file__) + "/static/")



import warnings
from typing import List

from django.http import HttpRequest, JsonResponse
from django.views import View

from .actions import BaseAction
from .types import DataFile


warnings.warn("This module is deprecated. Please use the new signing API", DeprecationWarning)


class GenericDigitalSignViewMixin(object):
    request: HttpRequest

    def get_files(self) -> List[DataFile]:
        """
        This should be implemented on view level, and should return a list of files that should be digitally signed.

        Will be ignored if `get_bdoc_container_file()` returns not None
        """
        return []

    def get_bdoc_container_file(self) -> str:
        """
        Returns path to a container that should be digitally signed.

        If there is a container, then get_files() will be ignored, and a new signature will be added to the container.
        """
        pass


class ApiView(View):
    def dispatch(self, request, *args, **kwargs):
        response = super().dispatch(request, *args, **kwargs)
        if isinstance(response, (dict, list, tuple)):
            return JsonResponse(response)

        return response


class SignStatusViewMixin:
    """Gets the status of signing and finalizes the process if necessary."""

    ACTION_CLASS: BaseAction

    def build_action_kwargs(self, request):
        return {}

    def post(self, request, *args, **kwargs):
        action_kwargs = self.build_action_kwargs(request)
        return self.ACTION_CLASS.do_action(self, **action_kwargs)

    def get(self, request, *args, **kwargs):
        return self.post(request, *args, **kwargs)



import re
from collections import namedtuple
from typing import Optional

from esteid import settings
from esteid.constants import Languages
from esteid.exceptions import InvalidIdCode, InvalidParameter
from esteid.signing.types import InterimSessionData
from esteid.types import PredictableDict
from esteid.validators import id_code_ee_is_valid


PHONE_NUMBER_REGEXP = settings.MOBILE_ID_PHONE_NUMBER_REGEXP

AuthenticateResult = namedtuple(
    "AuthenticateResult",
    [
        "session_id",
        "hash_type",
        "hash_value",
        "verification_code",
        "hash_value_b64",
    ],
)

AuthenticateStatusResult = namedtuple(
    "AuthenticateStatusResult",
    [
        "certificate",  # DER-encoded certificate
        "certificate_b64",  # Base64-encoded DER-encoded certificate
    ],
)

SignResult = namedtuple(
    "SignResult",
    [
        "session_id",
        "digest",
        "verification_code",
    ],
)

# Note: MobileID doesn't return a certificate for SignStatus. It is set from a previous call to `/certificate`
SignStatusResult = namedtuple(
    "SignStatusResult",
    [
        "signature",
        "signature_algorithm",
        "certificate",
    ],
)


class UserInput(PredictableDict):
    phone_number: str
    id_code: str
    language: Optional[str]

    def is_valid(self, raise_exception=True):
        result = super().is_valid(raise_exception=raise_exception)
        if result:
            if not self.phone_number or PHONE_NUMBER_REGEXP and not re.match(PHONE_NUMBER_REGEXP, self.phone_number):
                if not raise_exception:
                    return False
                raise InvalidParameter(param="phone_number")
            if not id_code_ee_is_valid(self.id_code):
                if not raise_exception:
                    return False
                raise InvalidIdCode
            if not (self.get("language") and self.language in Languages.ALL):
                self.language = settings.MOBILE_ID_DEFAULT_LANGUAGE
        return result


class MobileIdSessionData(InterimSessionData):
    session_id: str



__all__ = [
    "Container",
    "DataFile",
    "SignViewDjangoMixin",
    "SignViewMixin",
    "SignViewRestMixin",
    "Signer",
]

from pyasice import Container

from .signer import Signer
from .types import DataFile
from .views import SignViewDjangoMixin, SignViewMixin, SignViewRestMixin



from esteid.types import PredictableDict


class SessionData(PredictableDict):
    """
    Wrapper for temporary data stored between authentication polling requests.

    Contains the upstream service's session ID and the hash value used to calculate the authentication signature,
     along with a timestamp that is used to determine session validity timeout.
    """

    timestamp: int
    session_id: str
    hash_value_b64: str


class AuthenticationResult(PredictableDict):
    country: str
    id_code: str
    given_name: str
    surname: str
    certificate_b64: str



import base64
import typing
from typing import Literal

import pyasice

from ..constants import HASH_SHA256, HASH_SHA384, HASH_SHA512
from ..exceptions import InvalidParameters, SignatureVerificationError
from ..types import PredictableDict
from ..util import generate_hash


SignatureAlgorithm = Literal["ES256", "ES384", "ES512", "PS256", "PS384", "PS512", "RS256", "RS384", "RS512"]
SIGNATURE_ALGORITHMS = list(typing.get_args(SignatureAlgorithm))


def require_not_empty(value, param):
    if not value:
        raise InvalidParameters(f"Parameter {param} is required")


HASH_ALGORITHM_MAPPING = {
    "ES256": HASH_SHA256,
    "ES384": HASH_SHA384,
    "ES512": HASH_SHA512,
    "PS256": HASH_SHA256,
    "PS384": HASH_SHA384,
    "PS512": HASH_SHA512,
    "RS256": HASH_SHA256,
    "RS384": HASH_SHA384,
    "RS512": HASH_SHA512,
}


class LibraryAuthenticateResponse(PredictableDict):
    """
    Response to the web-eid `authenticate` call.

    Ref: https://github.com/web-eid/web-eid.js#authenticate-result
    """

    # base64-encoded DER encoded authentication certificate of the user
    unverifiedCertificate: str

    # algorithm used to produce the authentication signature
    algorithm: SignatureAlgorithm

    # base64-encoded signature of the token
    signature: str

    # type identifier and version of the token format separated by a colon character.
    #  example "web-eid:1.0"
    format: str

    # URL identifying the name and version of the application that issued the token
    #  example "https://web-eid.eu/web-eid-app/releases/2.0.0+0"
    appVersion: str

    def is_valid(self, raise_exception=True):
        res = super().is_valid(raise_exception)

        if res:
            require_not_empty(self.algorithm, "algorithm")
            require_not_empty(self.signature, "signature")

            if self.algorithm not in SIGNATURE_ALGORITHMS:
                raise InvalidParameters("Unsupported signature algorithm")

        return res

    @property
    def certificate_bytes(self):
        return base64.b64decode(self.unverifiedCertificate)

    @property
    def signature_bytes(self):
        return base64.b64decode(self.signature)

    def validate_signature(self, certificate, origin: str, challenge_nonce: bytes):
        """
        Validates that the signature of the authentication token is valid
        """
        require_not_empty(challenge_nonce, "challenge_nonce")

        hash_algorithm = HASH_ALGORITHM_MAPPING[self.algorithm]

        origin_hash = generate_hash(hash_algorithm, origin.encode())
        nonce_hash = generate_hash(hash_algorithm, challenge_nonce)
        hash_value = origin_hash + nonce_hash

        try:
            pyasice.verify(certificate, self.signature_bytes, hash_value, hash_algorithm, prehashed=False)
        except pyasice.SignatureVerificationError as e:
            raise SignatureVerificationError from e



import logging
import os
from tempfile import NamedTemporaryFile
from time import time
from typing import Dict, List, Type

from esteid_certificates import get_certificate

import pyasice
from pyasice import Container, XmlSignature

from esteid import settings
from esteid.exceptions import EsteidError, SigningSessionDoesNotExist, SigningSessionExists

from .types import DataFile, InterimSessionData


logger = logging.getLogger(__name__)


class Signer:
    """
    Abstract pluggable signer.

    Receives initial parameters from a view (or any other caller), based on which
    it selects the concrete Signer subclass.

    The concrete Signer:
    * does the necessary preparation of container/signature
      for the files provided by the view;
    * calls the associated services as necessary, to obtain the signature;
    * finalizes the container.

    The concrete Signer is also associated with an initial parameters type,
    a session storage type, ... - akin to DRF serializers.

    The only requirement to a concrete signer class is that it
    inherit from the Signer class and register itself
    by being imported. The registration is done by `__init_subclass__()`.

    The signer class' name, lowercase, without `signer` suffix, will be the signing method name.
    """

    SIGNING_METHODS: Dict[str, Type["Signer"]] = {}

    # The session data class, configurable to match the necessary session data profile
    SessionData = InterimSessionData

    # Instance variable that holds session data
    session_data: InterimSessionData

    # This can be shared by subclasses, unless it's desired to be able to sign by different methods at once
    # (which is probably useless from the user point of view)
    _SESSION_KEY = f"{__name__}.session"

    # timeout in seconds, after which a fresh session can be started even if old session data is present.
    SESSION_VALIDITY_TIMEOUT = 60 * 2

    # the signing party's ID code, public attribute/property for use in checks etc.
    id_code: str

    # Abstract Methods

    def prepare(self, container: Container = None, files: List[DataFile] = None) -> dict:
        """
        Abstract method. Prepares the container, either from an existing one or from files.

        Returns a dict of data to display to user, if necessary: e.g. verification code.

        Preparing the container means creating a temporary container file, and a temporary XML signature,
        which requires a user certificate. Getting this certificate is the responsibility of the Signer class.

        Note 1: if `container` exists, it should be used regardless of whether `files` are also passed.
        Note 2: if `container` exists, it should be copied to a temporary file, so that if during the signing process
                the original container file gets modified, we would still be signing the original one.

        See also the `open_container()` method.
        """
        raise NotImplementedError

    def finalize(self, data=None) -> Container:
        """
        Abstract method. Checks the signing process status and finalizes the container when complete.

        Can accept data as necessary (e.g. when signing with ID card, it receives `signature_value`).
        Returns a handle to the final container. This can be e.g. an open file, or BytesIO.
        Raises Pending, Canceled, etc.
        """
        raise NotImplementedError

    # Customizable methods

    def setup(self, initial_data: dict = None):
        """Customize this to receive and check any data prior to `prepare()`"""
        pass

    def save_session_data(self, *, digest: bytes, container: Container, xml_sig: XmlSignature):
        """
        Saves the interim session data along with a timestamp that is used to determine session validity.

        Can be extended to accept additional arguments
        """
        data_obj = self.session_data

        data_obj.digest = digest
        data_obj.timestamp = int(time())

        with NamedTemporaryFile(delete=False) as temp_signature_file:
            temp_signature_file.write(xml_sig.dump())
        data_obj.temp_signature_file = temp_signature_file.name

        with NamedTemporaryFile("wb", delete=False) as temp_container_file:
            temp_container_file.write(container.finalize().getbuffer())
        data_obj.temp_container_file = temp_container_file.name

        self.session[self._SESSION_KEY] = dict(data_obj)

    # Methods that probably do not need overriding

    def load_session_data(self, session) -> InterimSessionData:
        try:
            session_data = session[self._SESSION_KEY]
        except KeyError:
            session_data = {}

        try:
            session_data = self.SessionData(session_data)
        except TypeError:
            session_data = self.SessionData()
            self._cleanup_session(session)

        # Not doing session data validation here, because
        # an instance of another type may need different data
        return session_data

    def __init__(self, session, initial=False):
        """
        Initializes the necessary session data.

        Takes a session object, e.g. django request.session,
        and a flag that tells whether to start new session or attempt to load an existing one
        """
        session_data = self.load_session_data(session)
        if initial:
            if session_data:
                try:
                    timestamp = session_data.timestamp
                except AttributeError:
                    timestamp = 0

                if time() < timestamp + self.SESSION_VALIDITY_TIMEOUT:
                    raise SigningSessionExists("Another signing session already in progress")

                # session expired => create a fresh data store
                session_data = self.SessionData()

                # clear the old session data. This incurs no DB overhead:
                # Django issues the actual DB query only in the process_response phase.
                self._cleanup_session(session)
        else:
            if not session_data:
                raise SigningSessionDoesNotExist("No active signing session found")

            try:
                session_data.is_valid()
            except ValueError as e:
                raise SigningSessionDoesNotExist("Invalid signing session") from e

            if time() > session_data.timestamp + self.SESSION_VALIDITY_TIMEOUT:
                raise SigningSessionDoesNotExist("This signing session has expired")

        self.session = session
        self.session_data = session_data

    def cleanup(self):
        """
        Cleans temporary signing session data and files.
        """
        return self._cleanup_session(self.session)

    @classmethod
    def start_session(cls, session, initial_data) -> "Signer":
        """
        Initializes a fresh signing session.
        """
        signer = cls(session, initial=True)
        signer.setup(initial_data)
        return signer

    @classmethod
    def load_session(cls, session) -> "Signer":
        """
        Continues (loads) an existing signing session from the `session` object
        """
        return cls(session, initial=False)

    @classmethod
    def _cleanup_session(cls, session):
        data = session.pop(cls._SESSION_KEY, None)
        if not data:
            return

        try:
            data = cls.SessionData(data)
        except TypeError:
            return

        try:
            os.remove(data.temp_container_file)
        except (AttributeError, FileNotFoundError):
            pass

        try:
            os.remove(data.temp_signature_file)
        except (AttributeError, FileNotFoundError):
            pass

    @staticmethod
    def open_container(container: Container = None, files: List[DataFile] = None) -> Container:
        if container:
            if not isinstance(container, Container):
                raise ValueError(
                    f"Expected container to be a pyasice.Container instance, got {container.__class__.__name__}"
                )
        elif files:
            container = Container()
            for f in files:
                container.add_file(f.file_name, f.read(), f.mime_type)
        else:
            raise ValueError("Either container path or list of files must be present and not empty")
        return container

    @staticmethod
    def finalize_xml_signature(xml_sig: XmlSignature):
        """
        Makes requests to OCSP and Time stamping services and embeds the responses in the XML signature.
        """
        issuer_cert = get_certificate(xml_sig.get_certificate_issuer_common_name())

        try:
            pyasice.finalize_signature(
                xml_sig,
                issuer_cert,
                lt_ts=settings.ESTEID_USE_LT_TS,
                ocsp_url=settings.OCSP_URL,
                tsa_url=settings.TSA_URL,
            )
        except pyasice.Error as e:
            logger.exception("Signature confirmation error")
            raise EsteidError("Signature confirmation error") from e

    # "Magic" registration of subclasses

    @staticmethod
    def select_signer(signing_method: str) -> Type["Signer"]:
        try:
            signer_class = Signer.SIGNING_METHODS[signing_method]
        except KeyError as e:
            raise EsteidError(f"Failed to load signer: method `{signing_method}` not registered") from e
        return signer_class

    @classmethod
    def get_method_name(cls):
        method = cls.__name__.lower()
        if method.endswith("signer"):
            method = method[:-6]
        return method

    def __init_subclass__(cls):
        """Registers subclasses automatically"""
        method = cls.get_method_name()
        assert method not in Signer.SIGNING_METHODS, f"A Signer for {method} is already registered"

        Signer.SIGNING_METHODS[method] = cls



import logging
from typing import BinaryIO, Type, TYPE_CHECKING, Union

from django.http import HttpRequest, JsonResponse

import pyasice

from esteid import settings
from esteid.exceptions import ActionInProgress, AlreadySignedByUser
from esteid.mixins import DjangoRestCompatibilityMixin, SessionViewMixin
from esteid.types import Signer as SignerData

from .signer import Signer


try:
    from rest_framework.exceptions import ValidationError as DRFValidationError
except ImportError:
    # If rest framework is not installed, create a stub class so the isinstance check is always false
    class DRFValidationError:
        pass


if TYPE_CHECKING:
    # Make type checkers aware of request.session attribute which is missing on the HttpRequest class
    from django.contrib.sessions import base_session

    class RequestType(HttpRequest):
        session: base_session.AbstractBaseSession
        data: dict


logger = logging.getLogger(__name__)


class SignViewMixin(SessionViewMixin):
    # these come from the `url()` definition as in `View.as_view(signing_method='...')`, either one is enough
    signer_class: Type[Signer] = None
    signing_method: str = None

    def get_container(self, *args, **kwargs) -> Union[str, BinaryIO, pyasice.Container]:
        """
        Returns [path to|file handle of] the container to sign, if it exists prior to signing

        A string value is treated as a file path.
        A file handle is anything that provides a `read()` method that returns bytes.
        """
        raise NotImplementedError

    def save_container(self, container: pyasice.Container, *args, **kwargs):
        """
        Receives a container instance, expected to save the file contents as appropriate.

        Example with Django's UploadedFile:

            from esteid.compat import container_info
            instance = self.get_object()
            # Be sure to call `container_info(container)` before `container.finalize()`
            instance.container_info = container_info(container)
            instance.container = UploadedFile(container.finalize(), "signed_document.doc", container.MIME_TYPE)
            # OR:
            # buffer = container.finalize().getbuffer()
            # instance.container = SimpleUploadedFile("signed_document.doc", buffer, container.MIME_TYPE)
            instance.save()

        """
        raise NotImplementedError

    def get_files_to_sign(self, *args, **kwargs) -> list:
        """
        Returns list of files to sign, unless there is a pre-created container
        """
        raise NotImplementedError

    def get_success_response(self, *args, **kwargs):
        """
        Generates a success response when the signing process is complete (in `finalize()`)

        Can be customized to return additional data, such as links to download the container
        """
        return JsonResponse({"status": self.Status.SUCCESS})

    def check_eligibility(self, signer: Signer, container: pyasice.Container = None):
        """
        Performs a check whether a signing party is eligible to sign the container.

        Override it in subclasses as necessary:
        """
        if container is None or settings.ESTEID_ALLOW_ONE_PARTY_SIGN_TWICE:
            return

        signatories = []
        for signature in container.iter_signatures():
            subject_cert = signature.get_certificate()
            signatory = SignerData.from_certificate(subject_cert)
            signatories.append(signatory.id_code)

        if signer.id_code in signatories:
            raise AlreadySignedByUser

    def select_signer_class(self) -> Type["Signer"]:
        if self.signer_class is not None:
            return self.signer_class
        return Signer.select_signer(self.signing_method)

    def start_session(self, request: "RequestType", *args, **kwargs):
        """
        Initiates a signing session
        """

        signer_class = self.select_signer_class()
        signer = signer_class.start_session(request.session, request.data)

        try:
            container = self.get_container(*args, **kwargs)
        except NotImplementedError:
            container = None

        if not container:
            files_to_sign = self.get_files_to_sign(*args, **kwargs)
        else:
            files_to_sign = None
            if not isinstance(container, pyasice.Container):
                if isinstance(container, str):
                    container = pyasice.Container.open(container)
                else:
                    container = pyasice.Container(container)

        self.check_eligibility(signer, container)

        response_to_user = signer.prepare(container, files_to_sign)

        return JsonResponse({**response_to_user, "status": self.Status.SUCCESS})

    def finish_session(self, request: "RequestType", *args, **kwargs):
        """
        Checks the status of a signing session and attempts to finalize signing
        """
        signer_class = self.select_signer_class()
        signer = signer_class.load_session(request.session)

        do_cleanup = True

        try:
            container = signer.finalize(getattr(request, "data", None))
            self.save_container(container, *args, **kwargs)

        except ActionInProgress as e:
            do_cleanup = False
            return JsonResponse({"status": self.Status.PENDING, **e.data}, status=e.status)

        finally:
            if do_cleanup:
                signer.cleanup()

        return self.get_success_response(*args, **kwargs)


class SignViewRestMixin(SignViewMixin):
    """
    To be used with rest-framework's APIView.
    """


class SignViewDjangoMixin(DjangoRestCompatibilityMixin, SignViewMixin):
    """
    To be used with plain Django class-based views (No rest-framework).

    Adds `data` attribute to the request with the POST or JSON data.
    """



import logging
from typing import Type, TYPE_CHECKING

from django.http import HttpRequest, JsonResponse

from esteid.exceptions import ActionInProgress
from esteid.mixins import DjangoRestCompatibilityMixin, SessionViewMixin

from .authenticator import Authenticator
from .types import AuthenticationResult


try:
    from rest_framework.exceptions import ValidationError as DRFValidationError
except ImportError:
    # If rest framework is not installed, create a stub class so the isinstance check is always false
    class DRFValidationError:
        pass


if TYPE_CHECKING:
    # Make type checkers aware of request.session attribute which is missing on the HttpRequest class
    from django.contrib.sessions import base_session

    class RequestType(HttpRequest):
        session: base_session.AbstractBaseSession
        data: dict


logger = logging.getLogger(__name__)


def get_origin(request):
    origin = f"{request.scheme}://{request.get_host()}"

    return origin


class AuthenticationViewMixin(SessionViewMixin):
    # these come from the `url()` definition as in `View.as_view(authentication_method='...')`, either one is enough
    authentication_method: str = None
    authenticator: Type[Authenticator] = None

    def on_auth_success(self, request, data: AuthenticationResult):
        """
        A hook to make use of the authentication data once the process is complete.

        May be used to store the data into session, authenticate a user etc.
        """
        pass

    def success_response(self, request, data: AuthenticationResult):
        """Customizable response on success"""
        return JsonResponse({**data, "status": self.Status.SUCCESS})

    def select_authenticator_class(self) -> Type["Authenticator"]:
        if self.authenticator is not None:
            return self.authenticator
        return Authenticator.select_authenticator(self.authentication_method)

    def start_session(self, request: "RequestType", *args, **kwargs):
        """
        Initiates an authentication session.
        """

        auth_class = self.select_authenticator_class()
        authenticator = auth_class.start_session(request.session, request.data, origin=get_origin(request))

        do_cleanup = True

        try:
            result = authenticator.authenticate()

        except ActionInProgress as e:
            do_cleanup = False
            # return SUCCESS to indicate that the upstream service successfully accepted the request
            return JsonResponse({"status": self.Status.SUCCESS, **e.data}, status=e.status)

        else:
            # Handle a theoretical case of immediate authentication
            self.on_auth_success(request, result)
            return JsonResponse({**result, "status": self.Status.SUCCESS})

        finally:
            if do_cleanup:
                authenticator.cleanup()

    def finish_session(self, request: "RequestType", *args, **kwargs):
        """
        Checks the status of an authentication session
        """
        authenticator_class = self.select_authenticator_class()
        authenticator = authenticator_class.load_session(request.session, origin=get_origin(request))

        do_cleanup = True

        try:
            result = authenticator.poll(request.data)

        except ActionInProgress as e:
            do_cleanup = False
            return JsonResponse({"status": self.Status.PENDING, **e.data}, status=e.status)

        else:
            self.on_auth_success(request, result)
            return self.success_response(request, result)

        finally:
            if do_cleanup:
                authenticator.cleanup()

    def handle_delete_request(self, request):
        authenticator_class = self.select_authenticator_class()
        authenticator = authenticator_class.load_session(request.session, origin=get_origin(request))

        authenticator.cleanup()



class AuthenticationViewRestMixin(AuthenticationViewMixin):
    """
    To be used with rest-framework's APIView.
    """
    def delete(self, request, *args, **kwargs):
        self.handle_delete_request(request)

        return JsonResponse({"status": self.Status.CANCELLED})


class AuthenticationViewDjangoMixin(DjangoRestCompatibilityMixin, AuthenticationViewMixin):
    """
    To be used with plain Django class-based views (No rest-framework).

    Adds `data` attribute to the request with the POST or JSON data.
    """
    def delete(self, request, *args, **kwargs):
        self.handle_delete_request(request)

        return JsonResponse({"status": self.Status.CANCELLED})



from collections import namedtuple
from typing import Optional

from esteid.constants import Countries
from esteid.signing.types import InterimSessionData
from esteid.types import PredictableDict
from esteid.validators import validate_id_code


AuthenticateResult = namedtuple(
    "AuthenticateResult",
    [
        "session_id",
        "hash_type",
        "hash_value",
        "hash_value_b64",
        "verification_code",
    ],
)

AuthenticateStatusResult = namedtuple(
    "AuthenticateStatusResult",
    [
        "document_number",
        "certificate",  # DER-encoded certificate
        "certificate_b64",  # Base64-encoded DER-encoded certificate
        "certificate_level",
    ],
)

SignResult = namedtuple(
    "SignResult",
    [
        "session_id",
        "digest",
        "verification_code",
    ],
)

SignStatusResult = namedtuple(
    "SignStatusResult",
    [
        "document_number",
        "signature",
        "signature_algorithm",
        "certificate",  # DER-encoded certificate
        "certificate_level",
    ],
)


class UserInput(PredictableDict):
    id_code: str
    country: Optional[str]

    def is_valid(self, raise_exception=True):
        """
        Raises InvalidIdCode or ValueError
        """
        result = super().is_valid(raise_exception=raise_exception)
        if result:
            if not (self.get("country") and self.country in Countries.ALL):
                self.country = Countries.ESTONIA

            validate_id_code(self.id_code, self.country)

        return result


class SmartIdSessionData(InterimSessionData):
    session_id: str



import base64
import logging
from time import sleep, time

import requests

import pyasice

from ..base_service import BaseSKService
from ..constants import HASH_ALGORITHMS, HASH_SHA256
from ..exceptions import ActionInProgress, CanceledByUser
from ..exceptions import EsteidError as SmartIDError
from ..exceptions import (
    PermissionDenied,
    SignatureVerificationError,
    UpstreamServiceError,
    UserNotRegistered,
    UserTimeout,
)
from ..util import generate_hash, secure_random
from ..validators import validate_id_code
from .constants import CERTIFICATE_LEVEL_QUALIFIED, CERTIFICATE_LEVELS, EndResults
from .types import AuthenticateResult, AuthenticateStatusResult, SignResult, SignStatusResult
from .utils import get_verification_code


logger = logging.getLogger(__name__)


class SmartIDService(BaseSKService):
    """Smart-ID Authentication and signing

    Based on https://github.com/SK-EID/smart-id-documentation

    Test api root: https://sid.demo.sk.ee/smart-id-rp/v1
    Production api root: https://rp-api.smart-id.com/v1

    Fixme:
        - Add more logging
    """

    DISPLAY_TEXT_AUTH = "Authenticate"
    DISPLAY_TEXT_SIGN = "Sign"

    NAME = "Smart ID"

    class Actions:
        AUTH = "/authentication/etsi/PNO{country}-{id_code}"
        SIGN_BY_DOCUMENT = "/signature/document/{document}"
        SELECT_CERTIFICATE = "/certificatechoice/etsi/PNO{country}-{id_code}"
        SELECT_CERTIFICATE_BY_DOCUMENT = "/certificatechoice/document/{document_number}"
        SESSION_STATUS = "/session/{session_id}"

    def authenticate(
        self, id_code, country, certificate_level=CERTIFICATE_LEVEL_QUALIFIED, message=None, hash_type=HASH_SHA256
    ):
        """Initiate an authentication session

        see https://github.com/SK-EID/smart-id-documentation#44-authentication-session

        :param str id_code: National identity number
        :param str country: Country as an uppercase ISO 3166-1 alpha-2 code (choices: Countries.ALL)
        :param str certificate_level: Level of certificate requested (choices: CERTIFICATE_LEVELS)
        :param str message: Text to display for authentication consent dialog on the mobile device
        :param str hash_type: Hash algorithm to use when generating a random hash value (choices: HASH_ALGORITHMS)
        :return AuthenticateResult: Result of the request
        """
        # Ensure required values are set
        validate_id_code(id_code, country)
        assert certificate_level in CERTIFICATE_LEVELS
        assert hash_type in HASH_ALGORITHMS

        random_bytes = secure_random(64)
        hash_value = generate_hash(hash_type, random_bytes)
        hash_value_b64 = base64.b64encode(hash_value).decode()

        endpoint = self.Actions.AUTH.format(country=country.upper(), id_code=id_code)

        data = {
            "certificateLevel": certificate_level,
            "hashType": hash_type,
            "hash": hash_value_b64,
            # Casting to str to ensure translations are resolved
            "displayText": str(message or self.DISPLAY_TEXT_AUTH),
            # Don't use nonce to so we can rely on idempotent behaviour
            #
            # From the docs:
            #
            # Whenever a RP session creation request (POST to certificatechoice/, signature/, authentication/) is
            # repeated inside a given timeframe with exactly the same parameters, session ID of an existing
            # session can be returned as a result.
            #
            # This allows to retry RP POST requests in case of communication errors. Retry timeframe is 15 seconds.
            #
            # When requestor wants, it can override the idempotent behaviour inside of this timeframe using an
            # optional "nonce" parameter present for all POST requests. Normally, that parameter can be omitted.
            # 'nonce': None,
        }

        try:
            result = self.invoke(
                endpoint,
                method="POST",
                data=data,
            )

        except UpstreamServiceError as e:
            self._handle_specific_errors(e)
            raise

        return AuthenticateResult(
            session_id=result["sessionID"],
            hash_type=hash_type,
            hash_value=hash_value,
            hash_value_b64=hash_value_b64,
            verification_code=get_verification_code(hash_value),
        )

    def status(self, session_id, hash_value, timeout=10000):
        """Retrieve session result from Smart-ID backend

        see https://github.com/SK-EID/smart-id-documentation#46-session-status

        :param session_id: session ID, from I{authenticate} Result
        :param hash_value: hash value that was sent to I{authenticate}
        :param int timeout: Request long poll timeout value in milliseconds (Note: server uses a default
                         if client does not send it)
        :rtype: AuthenticateStatusResult
        """
        data = self._get_session_response(session_id, timeout)

        # result.documentNumber (str): Document number, can be used in further signature and authentication
        #                              requests to target the same device. Note: Only available if
        #                              result.endResult is END_RESULT_OK
        document_number = data["result"]["documentNumber"]

        # signature.value (str): Signature value, base64 encoded.
        signature_value = base64.b64decode(data["signature"]["value"])
        # signature.algorithm (str): Signature algorithm, in the form of sha256WithRSAEncryption
        signature_algorithm = data["signature"]["algorithm"]
        assert signature_algorithm[:6].upper() in HASH_ALGORITHMS

        # cert: Only available if result.endResult is OK
        # cert.value (str): Certificate value, DER+Base64 encoded
        cert_b64 = data["cert"]["value"]
        cert_value = base64.b64decode(cert_b64)

        # cert.certificateLevel (str): Level of Smart-ID certificate:
        #                              ADVANCED - Used for Smart-ID basic.
        #                              QUALIFIED - Used for Smart-ID.
        # Note: Not really sure how this affects the data inside the certificate
        certificate_level = data["cert"]["certificateLevel"]

        try:
            pyasice.verify(cert_value, signature_value, hash_value, signature_algorithm[:6], prehashed=True)
        except pyasice.SignatureVerificationError as e:
            raise SignatureVerificationError from e

        return AuthenticateStatusResult(
            document_number=document_number,
            certificate=cert_value,
            certificate_b64=cert_b64,
            certificate_level=certificate_level,
        )

    def sign_by_document_number(
        self,
        document_number: str,
        signed_data: bytes,
        certificate_level=CERTIFICATE_LEVEL_QUALIFIED,
        message=None,
        hash_type=HASH_SHA256,
    ) -> SignResult:
        """Initiate a signature session by document number.

        This method is preferred over signing by id_code/country, and requires a prior authentication to get the
          document number, and also selecting a signing certificate.

        see https://github.com/SK-EID/smart-id-documentation#2310-signing-session

        :param document_number: Document number, obtained from auth session
        :param signed_data: Binary data to sign
        :param certificate_level: Level of certificate requested (choices: CERTIFICATE_LEVELS)
        :param str message: Text to display for authentication consent dialog on the mobile device
        :param str hash_type: Hash algorithm used to sign data
        :return SignResult: Result of the request
        """
        # Ensure required values are set
        assert document_number
        assert signed_data
        assert certificate_level in CERTIFICATE_LEVELS
        assert hash_type
        hash_type = hash_type.upper()
        assert hash_type in HASH_ALGORITHMS

        content_hash = generate_hash(hash_type, signed_data)
        hash_value_b64 = base64.b64encode(content_hash)

        endpoint = self.Actions.SIGN_BY_DOCUMENT.format(document=document_number)

        try:
            result = self.invoke(
                endpoint,
                method="POST",
                data={
                    "certificateLevel": certificate_level,
                    "hashType": hash_type,
                    "hash": hash_value_b64.decode("ascii"),
                    # Casting to str to ensure translations are resolved
                    "displayText": str(message or self.DISPLAY_TEXT_SIGN),
                },
            )

        except UpstreamServiceError as e:
            self._handle_specific_errors(e)
            raise

        return SignResult(
            session_id=result["sessionID"],
            digest=content_hash,
            verification_code=get_verification_code(content_hash),  # YES we hash the hash.
        )

    def sign_status(self, session_id, digest: bytes, timeout: int = 1000):
        """Retrieve signing session result from Smart-ID backend

        see https://github.com/SK-EID/smart-id-documentation#46-session-status

        :param session_id: session ID from I{sign} Result
        :param digest: the hash of the signed data
        :param int timeout: Request long poll timeout value in milliseconds (Note: server uses a default
                         if client does not send it)
        :rtype: SignStatusResult
        """
        data = self._get_session_response(session_id, timeout)

        cert_value = base64.b64decode(data["cert"]["value"])
        signature = base64.b64decode(data["signature"]["value"])
        signature_algorithm = data["signature"]["algorithm"]
        assert signature_algorithm[:6].upper() in HASH_ALGORITHMS

        pyasice.verify(cert_value, signature, digest, signature_algorithm[:6], prehashed=True)

        return SignStatusResult(
            document_number=data["result"]["documentNumber"],
            signature=signature,
            signature_algorithm=data["signature"]["algorithm"],
            certificate=cert_value,
            certificate_level=data["cert"]["certificateLevel"],
        )

    def select_signing_certificate(self, id_code, country, certificate_level=CERTIFICATE_LEVEL_QUALIFIED) -> tuple:
        """Obtain a certificate that will be used for signing.

        This method is REQUIRED prior to signing with `sign()`. Otherwise it's possible that `authenticate()` would
          return a different cert, and the XAdES signature would not be valid.

        :param id_code:
        :param country:
        :param certificate_level:
        :return: tuple[bytes, str] - ASN.1 (DER) certificate and document number
        """
        validate_id_code(id_code, country)
        endpoint = self.Actions.SELECT_CERTIFICATE.format(country=country.upper(), id_code=id_code)

        result = self.invoke(
            endpoint,
            method="POST",
            data={
                "certificateLevel": certificate_level,
            },
        )

        session_id = result["sessionID"]

        start_time = int(time())
        while True:
            # Note: Normally the session response is returned quickly to the first polling request.
            # In case this doesn't happen, we allow polling for 60 seconds with a 3-second interval.
            # These numbers are arbitrary but since they only serve as contingency fallback,
            # making constants out of them seems redundant.
            try:
                data = self._get_session_response(session_id, 1000)
                break
            except ActionInProgress as e:
                now = int(time())
                if now > start_time + 60:
                    raise UpstreamServiceError(
                        "Failed to get a certificate response within 60 seconds",
                        service=self.NAME,  # pylint: disable=no-member
                    ) from e
                sleep(3)

        document_number = data["result"]["documentNumber"]
        certificate = base64.b64decode(data["cert"]["value"])
        return certificate, document_number

    # ============
    # Internals
    # ============

    def _get_session_response(self, session_id, timeout):
        """Perform a request to session status.

        :param session_id:
        :param timeout:
        :return:
        """
        data = self.poll_session(session_id, endpoint_url=self.Actions.SESSION_STATUS, timeout=timeout)

        # result.endResult (str): End result of the transaction.
        # This structure is different from Mobile ID so can't efficiently put it into BaseSKService.poll_session()
        end_result = data["result"]["endResult"]

        if end_result != EndResults.OK:
            if end_result == EndResults.TIMEOUT:
                raise UserTimeout
            if end_result == EndResults.USER_REFUSED:
                raise CanceledByUser

            # Fail hard, if endResult is something unknown to us
            if end_result not in EndResults.ALL:
                raise SmartIDError(f"Unexpected end result {end_result}")

            raise UpstreamServiceError(end_result, service=self.NAME)  # pylint: disable=no-member

        return data

    def _handle_specific_errors(self, exc: UpstreamServiceError):
        """
        Handles some SmartID-specific errors
        """
        cause = getattr(exc, "__cause__", None)
        if isinstance(cause, requests.HTTPError):
            response = cause.response
            # From the docs:
            #
            # HTTP error code 403 - Relying Party has no permission to issue the request. This may happen when:
            #  Relying Party has no permission to invoke operations on accounts with ADVANCED certificates.
            if response.status_code == 403:
                raise PermissionDenied("Certificate level ADVANCED not allowed") from exc

            # From the docs:
            #
            # HTTP error code 404 - object described in URL was not found,
            # essentially meaning that the user does not have account in Smart-ID system.
            if response.status_code == 404:
                raise UserNotRegistered from exc

            # Log the HTTP response
            logger.exception(
                "The %s service returned an error %s at %s. Response:\n %s",
                self.NAME,  # pylint: disable=no-member
                response.status_code,
                response.url,
                response.text,
            )



import logging
from time import time
from typing import Dict, Optional, Type

from esteid.authentication.types import AuthenticationResult, SessionData
from esteid.exceptions import EsteidError, SigningSessionDoesNotExist, SigningSessionExists


logger = logging.getLogger(__name__)


class Authenticator:
    """
    Abstract pluggable authenticator.

    Receives initial parameters from a view (or any other caller), based on which
    it selects the concrete Authenticator subclass.

    The concrete Authenticator calls the associated services as necessary,
      to obtain the authentication data.

    The concrete Authenticator is also associated with an initial parameters type,
    a session storage type, ... - akin to DRF serializers.

    The only requirement to a concrete authenticator class is that it
    inherit from the Authenticator class and register itself
    by being imported. The registration is done by `__init_subclass__()`.

    The authenticator class' name, lowercase, without `authenticator` suffix, will be the authentication method name.
    """

    AUTHENTICATION_METHODS: Dict[str, Type["Authenticator"]] = {}

    _SESSION_KEY = f"{__name__}.session"

    # timeout in seconds, after which a fresh session can be started even if old session data is present.
    SESSION_VALIDITY_TIMEOUT = 60 * 2

    session_data: SessionData

    # Abstract methods

    def authenticate(self) -> Optional[AuthenticationResult]:
        """
        Initiate the authentication process.

        If the init request succeeded, raises an ActionInProgress error
        with the corresponding data such as verification code.

        Note: For a theoretical case when authentication happens immediately, we leave an opportunity
        that this method returns AuthenticationResult.
        """
        raise NotImplementedError

    def poll(self, initial_data: dict = None) -> AuthenticationResult:
        """
        Polls status of the authentication process.

        If authentication succeeded, returns an AuthenticationResult.
        Otherwise, raises an ActionInProgress error with the corresponding data such as verification code.
        """
        raise NotImplementedError

    # Customizable methods

    def setup(self, initial_data: dict = None):
        """Customize this to receive and check any data prior to `prepare()`"""
        pass

    # Session management.
    # NOTE: this part is largely a copy-paste from signing.

    def save_session_data(self, *, session_id, hash_value_b64):
        """
        Saves the session data between initialization and polling requests
        """
        session_data = self.session_data or SessionData()

        session_data.session_id = session_id
        session_data.hash_value_b64 = hash_value_b64
        session_data.timestamp = int(time())

        self.session[self._SESSION_KEY] = dict(session_data)

    def load_session_data(self, session) -> SessionData:
        try:
            session_data = session[self._SESSION_KEY]
        except KeyError:
            session_data = {}

        try:
            session_data = SessionData(session_data)
        except TypeError:
            session_data = SessionData()
            self._cleanup_session(session)

        # Not doing session data validation here, because
        # an instance of another type may need different data
        return session_data

    def __init__(self, session, initial=False, origin=None):
        """
        Initializes the necessary session data.

        Takes a session object, e.g. django request.session,
        and a flag that tells whether to start new session or attempt to load an existing one
        """
        session_data = self.load_session_data(session)
        if initial:
            if session_data:
                try:
                    timestamp = session_data.timestamp
                except AttributeError:
                    timestamp = 0

                if time() < timestamp + self.SESSION_VALIDITY_TIMEOUT:
                    raise SigningSessionExists("Another authentication session already in progress")

                # session expired => create a fresh data store
                session_data = SessionData()

                # wipe the old data from session.
                self._cleanup_session(session)
        else:
            if not session_data:
                raise SigningSessionDoesNotExist("No active authentication session found")

            try:
                session_data.is_valid()
            except ValueError as e:
                raise SigningSessionDoesNotExist("Invalid authentication session") from e

            if time() > session_data.timestamp + self.SESSION_VALIDITY_TIMEOUT:
                raise SigningSessionDoesNotExist("This authentication session has expired")

        self.session = session
        self.session_data = session_data
        self.origin = origin

    def cleanup(self):
        """
        Cleans temporary authentication session data and files.
        """
        return self._cleanup_session(self.session)

    @classmethod
    def start_session(cls, session, initial_data, origin) -> "Authenticator":
        """
        Initializes a fresh authentication session.
        """
        signer = cls(session, initial=True, origin=origin)
        signer.setup(initial_data)
        return signer

    @classmethod
    def load_session(cls, session, origin) -> "Authenticator":
        """
        Continues (loads) an existing authentication session from the `session` object
        """
        return cls(session, initial=False, origin=origin)

    @classmethod
    def _cleanup_session(cls, session):
        session.pop(cls._SESSION_KEY, None)

    # "Magic" registration of subclasses

    @staticmethod
    def select_authenticator(authentication_method: str) -> Type["Authenticator"]:
        try:
            authenticator_class = Authenticator.AUTHENTICATION_METHODS[authentication_method]
        except KeyError as e:
            raise EsteidError(f"Failed to load signer: method `{authentication_method}` not registered") from e
        return authenticator_class

    @classmethod
    def get_method_name(cls):
        method = cls.__name__.lower()
        if method.endswith("authenticator"):
            method = method[: -len("authenticator")]
        return method

    def __init_subclass__(cls):
        """Registers subclasses automatically"""
        method = cls.get_method_name()
        assert (
            method not in Authenticator.AUTHENTICATION_METHODS
        ), f"An Authenticator for {method} is already registered"

        Authenticator.AUTHENTICATION_METHODS[method] = cls



__all__ = ["IdCardAuthenticator", "IdCardSigner", "LibraryAuthenticateResponse"]

from .authenticator import IdCardAuthenticator
from .signer import IdCardSigner
from .types import LibraryAuthenticateResponse



# pragma: no cover
import base64
import logging
import warnings
from tempfile import NamedTemporaryFile

from django.http import HttpResponse, HttpResponseRedirect
from django.views.generic import TemplateView, View

from pyasice import Container

from .actions import (
    IdCardFinishAction,
    IdCardPrepareAction,
    MobileIdSignAction,
    MobileIdStatusAction,
    SmartIdSignAction,
    SmartIdStatusAction,
)
from .generic import ApiView, GenericDigitalSignViewMixin, SignStatusViewMixin
from .types import DataFile


warnings.warn("This module is deprecated. Please use the new API", DeprecationWarning)


logger = logging.getLogger(__name__)


class SKTestView(TemplateView):
    template_name = "esteid/test.html"

    def get_context_data(self, **kwargs):
        context = super(SKTestView, self).get_context_data(**kwargs)

        try:
            files = self.request.session["__ddoc_files"]

        except KeyError:
            files = {}

        context.update(
            {
                "files": files,
            }
        )

        return context

    def post(self, request, *args, **kwargs):
        """This method is only for testing and can
        be used to add some files to the session.

        actions:

            - add_file
            - remove_file
        """
        try:
            files = request.session["__ddoc_files"]

        except KeyError:
            files = {}

        action = request.POST.get("action", "add_file")

        if action == "remove_file":
            file_name = request.POST.get("file_name", "")

            n_files = {}
            for key, file in files.items():
                if key != file_name:
                    n_files[key] = file

            files = n_files

        else:
            for key, file in request.FILES.items():
                file_name = file.name

                idx = 1
                while file_name in files.keys():
                    file_name = f"{idx}_{file_name}"
                    idx += 1

                files[file_name] = {
                    "content": base64.b64encode(file.read()).decode(),
                    "content_type": file.content_type,
                    "size": file.size,
                }

        request.session["__ddoc_files"] = files

        return HttpResponseRedirect(".")


class TestFilesMixin(GenericDigitalSignViewMixin):
    def get_files(self):
        files = []

        for file_name, file in self.request.session["__ddoc_files"].items():
            files.append(
                DataFile(
                    file_name=file_name,
                    mimetype=file["content_type"],
                    content_type=file["content_type"],
                    size=file["size"],
                    content=base64.b64decode(file["content"]),
                )
            )

        return files


class TestSignStatusViewMixin(SignStatusViewMixin):
    """Gets the status of signing and finalizes the process if necessary."""

    def post(self, request, *args, **kwargs):
        result = super().post(request, *args, **kwargs)

        if result["success"]:
            container: Container = result.pop("container", None)

            if container:
                with NamedTemporaryFile("wb", delete=False) as f:
                    f.write(container.finalize().getbuffer())

                request.session["__ddoc_container_file"] = f.name
                logger.debug("Saved container to temp file %s", f.name)

        return result


class TestMobileIdSignView(TestFilesMixin, ApiView):
    def post(self, request, *args, **kwargs):
        return MobileIdSignAction.do_action(
            self,
            id_code=request.POST.get("id_code", ""),
            phone_number=request.POST.get("phone_nr", ""),
            language=request.POST.get("language", None),
        )


class TestMobileIdStatusView(TestSignStatusViewMixin, ApiView):
    ACTION_CLASS = MobileIdStatusAction


class TestSmartIdSignView(TestFilesMixin, ApiView):
    def post(self, request, *args, **kwargs):
        return SmartIdSignAction.do_action(
            self,
            id_code=request.POST.get("id_code", ""),
            language=request.POST.get("language", None),
        )


class TestSmartIdStatusView(TestSignStatusViewMixin, ApiView):
    ACTION_CLASS = SmartIdStatusAction


class TestDownloadContainerView(View):
    def get(self, request, *args, **kwargs):
        request.session.pop("__ddoc_files", None)

        container_file = request.session.pop("__ddoc_container_file", None)
        logging.info("Got container temp file name '%s'", container_file)
        file_contents = None
        if container_file:
            try:
                with open(container_file, "rb") as f:
                    file_contents = f.read()
            except FileNotFoundError:
                pass

        if not file_contents:
            response = HttpResponse("Error: no container file found", status=409)
        else:
            # Download the file
            response = HttpResponse(file_contents, content_type=Container.MIME_TYPE)
            response["Content-Disposition"] = "attachment; filename=" + "signed.bdoc"
        return response


class TestIdCardSignView(TestFilesMixin, ApiView):
    def post(self, request, *args, **kwargs):
        return IdCardPrepareAction.do_action(self, certificate=request.POST["certificate"])


class TestIdCardFinishView(TestSignStatusViewMixin, ApiView):
    ACTION_CLASS = IdCardFinishAction

    def build_action_kwargs(self, request):
        return {"signature_value": request.POST["signature_value"]}


class AuthenticationView(TemplateView):
    template_name = "esteid/authenticate.html"

    def __init__(self, *args, **kwargs):
        self.id_auth = None
        self.id_err = None
